from zope import component
from zope import interface

from zope.i18n.interfaces import ITranslationDomain
from zope.i18n.interfaces import INegotiator
from zope.i18n import interpolate
from zope.i18n.testmessagecatalog import TestMessageCatalog

from zope.app.i18n.translationdomain import TranslationDomain as ZopeTranslationDomain
from zope.app.i18n.interfaces import ILocalTranslationDomain

from Products.PlacelessTranslationService import PlacelessTranslationService
from Products.PlacelessTranslationService.GettextMessageCatalog import translationRegistry
from Products.PlacelessTranslationService.Negotiator \
     import Negotiator as PTSNegotiator

from Products.CMFPlone.interfaces import IPloneSiteRoot
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.interfaces import ISiteRoot

from interfaces import ICustomizableTranslationsTool, IListMessages

import sys
_pts = sys.modules['Products.PlacelessTranslationService.PlacelessTranslationService']
catalogRegistry = _pts.catalogRegistry

from utils import match, make_token, normalize_language_token

query_accuracy_threshold = 0.8

from OFS.SimpleItem import SimpleItem

class TranslationDomain(ZopeTranslationDomain, SimpleItem):
    def getDefaultLanguage(self):
        return getToolByName(self, 'portal_languages').getDefaultLanguage()

    def _getPTS(self):
        portal = getToolByName(self, 'portal_url').getPortalObject()
        return portal.aq_inner.aq_parent.Control_Panel.TranslationService
    
    def getCatalogs(self, language):
        pts = self._getPTS()

        for (ccode, cdomain), cnames in catalogRegistry.items():
            if (cdomain, ccode) == (self.domain, language):
                return (translationRegistry[pts._getOb(cname).getId()]._catalog
                        for cname in cnames)

        return []

    def _PTS_translate(self, msgid, mapping, context, target_language, default):
        pts = self._getPTS()
        return pts.translate(
            self.domain, msgid, mapping, context, target_language, default)

    def translate(self, msgid, mapping=None, context=None,
                  target_language=None, default=None):
        """This will try to translate using the local utility that we
        are, and then it'll try to fallback to PTS.
        """
        try:
            msgid = unicode(msgid)
        except UnicodeDecodeError:
            msgid = unicode(msgid, 'utf-8')

        text = super(TranslationDomain, self).translate(
            msgid, mapping, context, target_language, 0)

        if text:
            return text
        else:
            return self._PTS_translate(
                msgid, mapping, context, target_language, default)
            
class LocalTranslationDomainAdapter(object):
    component.adapts(IPloneSiteRoot)
    interface.implements(ICustomizableTranslationsTool)

    def __init__(self, portal):
        self.context = portal
        
    def addMessage(self, msgid, msg, domain, language):
        translationdomain = component.queryUtility(
            ILocalTranslationDomain, name=domain)
    
        # if this domain is not persistent, we're going to register a new
        # local translation domain
        if translationdomain is None:
            translationdomain = TranslationDomain()
            translationdomain.domain = domain
            
            sm = component.getSiteManager()
            assert sm is not component.getGlobalSiteManager(), (
                "Expected local site manager.")
            sm.registerUtility(ILocalTranslationDomain, translationdomain, name=str(domain))
        
        # add message to td
        translationdomain.addMessage(msgid, msg, language)

        # work-around for bug in zope.app.i18n.translationdomain
        translationdomain[language].domain = domain

    def removeMessage(self, msgid, domain, language):
        translationdomain = component.getUtility(ILocalTranslationDomain, name=domain)
        
        translationdomain.deleteMessage(msgid, language)
    
    def queryMessage(self, query, language, include_global=True):
        tokens = []
        result = []

        # get translation domains, local first
        tds = [td for name, td in component.getUtilitiesFor(ILocalTranslationDomain)]

        if include_global:
            tds += list(component.getAllUtilitiesRegisteredFor(ITranslationDomain))

            # create PTS translation domain objects on the fly
            for cat in translationRegistry.values():
                info = cat.info()
                try:
                    td = PTSTranslationDomain(info['domain'])
                    tds.append(td)
                except KeyError:
                    # catalog record broken; ignore.
                    pass
                
        for util in tds:
            messages = IListMessages(util)
            messages.context = self.context
            for message in messages.filter(language):
                # keep track of domain/msgid combinations (tokens) already added
                token = make_token(message['domain'], message['msgid'])
                if token in tokens:
                    continue

                msg = message['msgid'] + " " + message['msgstr']

                accuracy = match(msg, query)
            
                if accuracy > query_accuracy_threshold:
                    if token not in tokens:
                        result.append(message)
                        tokens.append(token)

        return result

    def listMessages(self, domain, language):
        translationdomain = component.getUtility(ILocalTranslationDomain, name=domain)

        return IListMessages(translationdomain)

class PTSTranslationDomain(object):
    def __init__(self, domain):
        self.domain = domain

class MessagesLister(object):
    interface.implements(IListMessages)

    def __init__(self, td):
        self.td = td
        
    def filter(self, language):
        token = normalize_language_token(language)
        return (message for message in self if \
                normalize_language_token(message['language']) == token)

class PTSTranslationDomainMessagesLister(MessagesLister):
    component.adapts(PTSTranslationDomain)

    def __iter__(self):
        portal = getToolByName(self.context, 'portal_url').getPortalObject()
        pts = portal.aq_inner.aq_parent.Control_Panel.TranslationService

        for (ccode, cdomain), cnames in catalogRegistry.items():
            if cdomain != self.td.domain:
                continue

            for cname in cnames:
                catalog = pts._getOb(cname)
                yield ccode, translationRegistry[catalog.getId()]._catalog
            

    def filter(self, language):
        token = normalize_language_token(language)
        mod_time = 0

        for lang, catalog in self:
            if normalize_language_token(lang) != token: continue

            for msgid, msgstr in catalog.items():
                if msgid == '': continue

                yield dict(msgid=msgid,
                           msgstr=msgstr,
                           domain=self.td.domain,
                           language=language,
                           mod_time=mod_time)

class GlobalTranslationDomainMessagesLister(MessagesLister):
    component.adapts(ITranslationDomain)
    
    def __iter__(self):
        for catalog in self.td._data.values():
            mod_time = 0
            if isinstance(catalog, TestMessageCatalog):
                continue
            for msgid, msgstr in catalog._catalog._catalog.items():
                if msgid == '':
                    continue

                yield dict(msgid=msgid,
                           msgstr=msgstr,
                           domain=self.td.domain,
                           language=catalog.language,
                           mod_time=mod_time)

class LocalTranslationDomainMessagesLister(MessagesLister):
    component.adapts(ILocalTranslationDomain)

    def __iter__(self):
        return (message for message in self.td.getMessages())

class Negotiator(object):
    interface.implements(INegotiator)
    
    def getLanguage(self, langs, env):
        # XXX: This is evil; we pretend that 'en' is always available;
        # this has the potential of making non human readable ids
        # appear.

        # The problem we're trying to solve here is that when you have
        # 'zh' as your second language, and 'en' as your first, we
        # would normally return the Chinese translation, which is
        # awkward.  Another workaround would be to provide an English
        # translation PO file.  This should be solved in a better way,
        # really.
        if 'en' not in langs:
            langs.insert(0, 'en')

        neg = PTSNegotiator()
        return neg.negotiate(langs, env, 'language')

negotiator = Negotiator()
